from collections import OrderedDict

import astropy.units as u
import numpy as np

from exorad.models.foregrounds.skyForegrounds import SkyForeground
from exorad.models.foregrounds.zodiacalForeground import ZodiacalFrg
from .task import Task


class EstimateZodi(Task):
    """
    It estimate the zodiacal radiance in the target direction for a specific wl range

    Parameters
    -----------
    zodi: dict
        zodiacal foreground description
    target: Target
        target class
    wl_range: (float, float)
        wavelength range to investigate. (wl_min, wl_max)

    Returns
    -------
    Target:
        updated target class
    """

    def __init__(self):
        self.addTaskParam('zodi', 'zodiacal foreground description')
        self.addTaskParam('target', 'target class')
        self.addTaskParam('wl_range', 'wavelength range to investigate')

    def execute(self):
        self.info('estimating zodiacal foreground')
        zodi_dict = self.get_task_param('zodi')
        target = self.get_task_param('target')
        wl_min, wl_max = self.get_task_param('wl_range')

        wl = np.logspace(np.log10((wl_min.to(u.um)).value),
                         np.log10((wl_max.to(u.um)).value), 6000) * u.um
        zodi = ZodiacalFrg(wl=wl, description=zodi_dict)

        if not hasattr(target, 'foreground'):
            setattr(target, 'foreground', OrderedDict())
        target.foreground['zodi'] = zodi.radiance
        self.set_output(target)


class EstimateForeground(Task):
    """
    It estimate the foreground radiance in the target direction for a specific wl range

    Parameters
    -----------
    foreground: dict
        foreground description
    target: Target
        target class
    wl_range: (float, float)
        wavelength range to investigate. (wl_min, wl_max)

    Returns
    -------
    Target:
        updated target class
    """

    def __init__(self):
        self.addTaskParam('foreground', 'foreground description')
        self.addTaskParam('target', 'target class')
        self.addTaskParam('wl_range', 'wavelength range to investigate')

    def execute(self):
        self.info('estimating custom foreground')
        foreground_dict = self.get_task_param('foreground')
        target = self.get_task_param('target')
        wl_min, wl_max = self.get_task_param('wl_range')

        wl = np.logspace(np.log10((wl_min.to(u.um)).value),
                         np.log10((wl_max.to(u.um)).value), 6000) * u.um
        foreground_name = foreground_dict['value']
        foreground = SkyForeground(wl, foreground_dict)
        if not hasattr(target, 'foreground'):
            setattr(target, 'foreground', OrderedDict())
        target.foreground[foreground_name] = foreground.skyFilter
        if not hasattr(target, 'skyTransmission'):
            from exorad.models.signal import Signal
            setattr(target, 'skyTransmission', Signal(wl, foreground.skyFilter.transmission))
        else:
            target.skyTransmission.data *= foreground.skyFilter.transmission
        self.set_output(target)


class EstimateForegrounds(Task):
    """
    It estimate the foreground radiance in the target direction for a specific wl range

    Parameters
    -----------
    foregrounds: dict
        foregrounds description
    target: Target
        target class
    wl_range: (float, float)
        wavelength range to investigate. (wl_min, wl_max)

    Returns
    -------
    Target:
        updated target class
    """

    def __init__(self):
        self.addTaskParam('foregrounds', 'foregrounds description')
        self.addTaskParam('target', 'target class')
        self.addTaskParam('wl_range', 'wavelength range to investigate')

    def execute(self):
        self.info('estimating foregrounds')
        target = self.get_task_param('target')
        foregrounds = self.get_task_param('foregrounds')
        wl_min, wl_max = self.get_task_param('wl_range')

        estimateZodi = EstimateZodi()
        estimateForeground = EstimateForeground()

        if isinstance(foregrounds, OrderedDict):
            for foreground in foregrounds:
                if foreground == 'zodiacal':
                    target = estimateZodi(zodi=foregrounds['zodiacal'],
                                          target=target,
                                          wl_range=(wl_min, wl_max))
                else:
                    target = estimateForeground(foreground=foregrounds[foreground],
                                                target=target,
                                                wl_range=(wl_min, wl_max))
        else:
            if foregrounds['value'] == 'zodiacal':
                target = estimateZodi(zodi=foregrounds,
                                      target=target,
                                      wl_range=(wl_min, wl_max))
            else:
                target = estimateForeground(foreground=foregrounds,
                                            target=target,
                                            wl_range=(wl_min, wl_max))
        self.set_output(target)



import logging
import os

import astropy.units as u
import h5py
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
from astropy.io.misc.hdf5 import read_table_hdf5

from exorad.log import Logger, setLogLevel

sns.set()

mpl_logger = logging.getLogger('matplotlib')
mpl_logger.setLevel(logging.WARNING)


class Plotter(Logger):
    """
        Plotter class. It offers a fast and easy way to produce diagnostic plots on the produced data.

        Parameters
        -----------
        input_table: Qtable
            table where to grab data and wl grid to plot.
        channels: dict
            dictionary describing the channels in the payload. Default is None
    """

    def __init__(self, input_table, channels=None, payload=None):
        self.set_log_name()
        self.inputTable = input_table
        self.channels = channels
        self.payload = payload
        self.fig = None
        self.fig_efficiency = None

    def plot_bands(self, ax, scale='log', channel_edges=True):
        """
        It plots the channels bands behind the indicated axes.

        Parameters
        -----------
        ax: matplotlib.axes
            axes where to plot the bands

        Returns
        --------
        matplotlib.axes.axes

        Note
        ----
        The Class input_table input parameter is required for this method to work.
        """
        channels = set(self.inputTable['chName'])
        palette = sns.color_palette('colorblind')
        tick_list = []

        for k, channelName in enumerate(channels):
            wl_min = min(self.inputTable['LeftBinEdge'][np.where(self.inputTable['chName'] == channelName)])
            if hasattr(wl_min,'unit'):
                wl_min = wl_min.value
            wl_max = max(self.inputTable['RightBinEdge'][np.where(self.inputTable['chName'] == channelName)])
            if hasattr(wl_max,'unit'):
                wl_max = wl_max.value
            ax.axvspan(wl_min, wl_max, alpha=0.1, zorder=0, color=palette[k])
            tick_list.append(wl_min)
        tick_list.append(wl_max)
        ax.set_xscale(scale)
        if channel_edges:
            ax.set_xticks(tick_list)
            ax.get_xaxis().set_major_formatter(matplotlib.ticker.ScalarFormatter())

        return ax

    def plot_efficiency(self, scale='log', channel_edges=True):
        """
        It produces a figure with payload efficiency over wavelength.
        The quantities reported are quantum efficiency, transmission and the photon conversion efficiency (pce)
        computed as the product of the quantum efficiency and transmission.

        Returns
        --------
        matplotlib.pyplot.figure

        Note
        ----
        The Class channels input parameter is required for this method to work.
        """
        if self.channels:
            from matplotlib.lines import Line2D
            from exorad.models.signal import Signal

            palette = sns.color_palette('bright')

            fig, ax = plt.subplots(1, 1, figsize=(10, 8))
            fig.suptitle('Payload photon conversion efficiency')

            if self.channels:
                keys = ['transmission', 'qe']
                for ch in self.channels:
                    pce = None
                    for i, key in enumerate(keys):
                        data = self.channels[ch].built_instr['{}_data'.format(key)]
                        sig = Signal(wl_grid=data['wl_grid']['value'] * u.Unit(data['wl_grid']['unit']),
                                     data=data['data']['value'])
                        ax.plot(sig.wl_grid, sig.data, color=palette[i], zorder=10)
                        if not pce:
                            pce = sig
                        else:
                            sig.spectral_rebin(pce.wl_grid)
                            pce.data *= sig.data
                    ax.plot(pce.wl_grid, pce.data, color=palette[i + 1], zorder=10)

                lines, labels = [], []
                for i, key in enumerate(keys):
                    lines.append(Line2D([0], [0], color=palette[i], lw=4))
                    labels.append(key)
                lines.append(Line2D([0], [0], color=palette[i + 1], lw=4))
                labels.append('pce')
            else:
                pce = self.inputTable['TR'] * self.inputTable['QE']
                self.inputTable['pce'] = pce
                keys = ['TR', 'QE', 'pce']
                self.debug('efficiency keys : {}'.format(keys))
                for e in keys:
                    ax.plot(self.inputTable['Wavelength'], self.inputTable[e], label=e, zorder=10)
                    # ax.plot(self.inputTable['Wavelength'], self.inputTable[e], c='None')
            ax.grid(zorder=0)
            ax = self.plot_bands(ax, scale, channel_edges)
            ax.legend(handles=lines, labels=labels)
            ax.set_title('Photon conversion efficiency')
            ax.set_xlabel('Wavelength [$\mu m$]')
            ax.set_ylabel('efficiency')
            # ax.set_xscale('log')

            self.fig_efficiency = fig
            return fig

        else:
            self.error('channels parameter is required for this method to work')

    def plot_noise(self, ax, ylim=None, scale='log', channel_edges=True):
        """
        It plots the noise components found in the input table in the indicated axes.

        Parameters
        -----------
        ax: matplotlib.axes
            axes where to plot the noises
        ylim: float
            if present, it sets the axes y bottom lim. Default is None.

        Returns
        --------
        matplotlib.axes.axes

        Note
        ----
        The Class input_table input parameter is required for this method to work.
        """
        palette = sns.color_palette('bright')

        noise_keys = [x for x in self.inputTable.keys() if 'noise' in x or 'custom' in x]
        self.debug('noise keys : {}'.format(noise_keys))
        for k, n in enumerate(noise_keys):
            if n == 'total_noise':
                ax.plot(self.inputTable['Wavelength'], self.inputTable[n], zorder=9, lw=1, c='k',
                        marker='.', markersize=5, label='total_noise', alpha=0.8)  # , c='None')
            else:
                if self.inputTable[n].unit == u.hr ** 0.5:
                    noise = self.inputTable[n]
                else:
                    self.debug('{} rescaled by starSignal_inAperture'.format(n))
                    noise = self.inputTable[n] / self.inputTable['star_signal_inAperture']
                # ax.scatter(self.inputTable['Wavelength'], noise, label=n, zorder=10, s=5, color=palette[k])
                ax.plot(self.inputTable['Wavelength'], noise, zorder=9, lw=1, alpha=0.5, marker='.',
                        label=n)  # color=palette[k])  # c='None')

        if not ylim:
            ax.set_ylim(1e-7)
        else:
            ax.set_ylim(ylim)
        #        ax.grid(zorder=0, which='both')
        locmaj = matplotlib.ticker.LogLocator(base=10, numticks=12)
        ax.yaxis.set_major_locator(locmaj)
        locmin = matplotlib.ticker.LogLocator(base=10.0, subs=(0.2, 0.4, 0.6, 0.8), numticks=12)
        ax.yaxis.set_minor_locator(locmin)
        ax.yaxis.set_minor_formatter(matplotlib.ticker.NullFormatter())
        ax.grid(axis='y', which='minor', alpha=0.3)
        ax.grid(axis='y', which='major', alpha=0.5)
        #        ax.legend(bbox_to_anchor=(1, 1))
        ax.legend(prop={'size': 12}, loc='upper left',
                  ncol=3, bbox_to_anchor=(0.05, -0.2), labelspacing=1.2, handlelength=1)
        ax.set_title('Noise Budget')
        ax.set_xlabel('Wavelength [$\mu m$]')
        ax.set_ylabel('relative noise [$\sqrt{{hr}}$]')
        ax.set_yscale('log')
        # ax.set_xscale('log')
        ax = self.plot_bands(ax, scale, channel_edges)
        return ax

    def plot_signal(self, ax, ylim=None, scale='log', channel_edges=True):
        """
        It plots the signal components found in the input table in the indicated axes.

        Parameters
        -----------
        ax: matplotlib.axes
            axes where to plot the signals
        ylim: float
            if present, it sets the axes y bottom lim. Default is None.

        Returns
        --------
        matplotlib.axes.axes

        Note
        ----
        The Class input_table input parameter is required for this method to work.
        """
        palette = sns.color_palette('bright')

        keys = [x for x in self.inputTable.keys() if 'signal' in x and 'noise' not in x]
        self.debug('signal keys : {}'.format(keys))
        for k, s in enumerate(keys):
            # ax.scatter(self.inputTable['Wavelength'], self.inputTable[s], label=s, zorder=10, s=5, color=palette[k])
            ax.plot(self.inputTable['Wavelength'], self.inputTable[s], zorder=9, lw=1, alpha=0.5, marker='.', label=s)
            # color=palette[k])  # , c='None')

        if not ylim:
            ax.set_ylim(1e-3)
        else:
            ax.set_ylim(ylim)
        #        ax.grid(zorder=0, which='both')
        locmaj = matplotlib.ticker.LogLocator(base=10, numticks=12)
        ax.yaxis.set_major_locator(locmaj)
        locmin = matplotlib.ticker.LogLocator(base=10.0, subs=(0.2, 0.4, 0.6, 0.8), numticks=12)
        ax.yaxis.set_minor_locator(locmin)
        ax.yaxis.set_minor_formatter(matplotlib.ticker.NullFormatter())
        ax.grid(axis='y', which='minor', alpha=0.3)
        ax.grid(axis='y', which='major', alpha=0.5)
        #        ax.legend(bbox_to_anchor=(1, 1))
        ax.legend(prop={'size': 12}, loc='upper left',
                  ncol=3, bbox_to_anchor=(0.05, -0.2), labelspacing=1.2, handlelength=1)
        ax.set_title('Signals')
        ax.set_xlabel('Wavelength [$\mu m$]')
        ax.set_ylabel('$ct/s$')
        ax.set_yscale('log')
        # ax.set_xscale('log')
        ax = self.plot_bands(ax, scale, channel_edges)

        return ax

    def plot_table(self, sig_ylim=None, noise_ylim=None, scale='log', channel_edges=True):
        """
        It produces a figure with signal and noise for the input table.

        Returns
        --------
        matplotlib.pyplot.figure
        (matplotlib.axes.axes, matplotlib.axes.axes)

        Note
        ----
        The Class input_table input parameter is required for this method to work.
        """
        fig, (ax1, ax2) = plt.subplots(2, 1, figsize=(10, 10))
        fig.suptitle(self.inputTable.meta['name'])
        ax1 = self.plot_signal(ax1, ylim=sig_ylim, scale=scale, channel_edges=channel_edges)
        ax2 = self.plot_noise(ax2, ylim=noise_ylim, scale=scale, channel_edges=channel_edges)
        plt.tight_layout()
        plt.subplots_adjust(top=0.9, bottom=0.22, hspace=0.7)
        self.fig = fig
        return fig, (ax1, ax2)

    def save_fig(self, name, efficiency=False):
        """
        It saves the produced figure.

        Parameters
        --------
        name: str
            figure name
        efficiency: bool
            if True it wll save the efficiency plot instead of the table plot. Default is False.
        """
        try:
            if efficiency:
                self.fig_efficiency.savefig('{}'.format(name))
            else:
                self.fig.savefig('{}'.format(name))

            self.info('plot saved in {}'.format(name))
        except AttributeError:
            self.error("the indicated figure is not available. Check if you have produced it.")

def main():
    import argparse
    from exorad.__version__ import __version__
    from exorad.utils.util import parse_range

    parser = argparse.ArgumentParser(description='ExoRad {}'.format(__version__))
    parser.add_argument("-i", "--input", dest='input', type=str,
                        required=True, help="Input h5 file to pass")
    parser.add_argument("-o", "--out", dest='out', type=str, default='None',
                        required=True, help="Output directory")
    parser.add_argument("-n", "--target-number", dest='target_number', type=str, default='all',
                        required=False, help="A list or range of targets to run")
    parser.add_argument("-t", "--target-name", dest='target_name', type=str, default='None',
                        required=False, help="name of the target to plot")
    parser.add_argument("-d", "--debug", dest='debug', default=False,
                        required=False, help="log output on screen", action='store_true')

    args = parser.parse_args()

    logger = logging.getLogger('exorad')
    from exorad.utils.ascii_art import ascii_plot
    logger.info(ascii_plot)
    logger.info('code version {}'.format(__version__))

    if args.debug: setLogLevel(logging.DEBUG)

    if not os.path.exists(args.out):
        os.makedirs(args.out)
        logger.info('output directory created')

    logger.info('reading {}'.format(args.input))
    file = h5py.File(args.input)

    if args.target_number != 'all' and args.target_name != 'None':
        logger.error('you cannot use both target number and target name')
        raise ValueError

    targets_dir = file['targets']
    targets_to_run_id = parse_range(args.target_number, len(targets_dir.keys()))
    targets_to_run = [list(targets_dir.keys())[n] for n in targets_to_run_id]

    if args.target_name != 'None':
        targets_to_run = [target for target in targets_to_run if target == args.target_name]

    for target in targets_to_run:
        target_dir = targets_dir[target]
        table_dir = target_dir['table']
        table = read_table_hdf5(table_dir, path='table')

        plotter = Plotter(input_table=table)
        plotter.plot_table()
        plotter.save_fig(os.path.join(args.out, '{}.png'.format(target)))
        plt.close()



from astropy.io.misc.hdf5 import read_table_hdf5

from exorad.models.instruments import Photometer, Spectrometer
from exorad.output.hdf5 import load
from .task import Task

instruments = {'photometer': Photometer,
               'spectrometer': Spectrometer}


class BuildInstrument(Task):
    """
    Initialize and build an instrument

    Parameters
    ----------
    type: str
        instrument class
    name: str
        instrument name
    description: dict
        instrument description dictionary
    payload: dict
        main payload. Default is None
    write: bool
        set to True to write the built dict to file. Default is None
    output: str
        output object

    Returns
    -------
    Instrument:
        return the built instrument class

    Raises
    ------
        KeyError
            if the indicated instrument class is not supported

    Examples
    --------
    >>> buildInstrument = BuildInstrument()
    >>> instrument = buildInstrument(type='photometer', name='Phot',
    >>>                              description=payload['channel']['Phot'],
    >>>                              payload=payload, write=False, output=None)
    """

    def __init__(self):
        self.addTaskParam('type', 'instrument type')
        self.addTaskParam('name', 'instrument name')
        self.addTaskParam('description', 'instrument description dictionary')
        self.addTaskParam('payload', 'main payload. Default is None')
        self.addTaskParam('write', 'write processed instrument to output file')
        self.addTaskParam('output', 'output object')

    def execute(self):
        try:
            instrumentClass = instruments[self.get_task_param('type')]
        except KeyError:
            self.error('invalid instrument class')
            raise ValueError
        instrument = instrumentClass(self.get_task_param('name'),
                                     self.get_task_param('description'),
                                     self.get_task_param('payload'))
        instrument.build()
        if self.get_task_param('write'):
            instrument.write(self.get_task_param('output'))
        self.set_output(instrument)


class BuildChannels(Task):
    """
    Initialize and build all the channels in the payload

    Parameters
    ----------
    payload: dict
        main payload. Default is None
    write: bool
        set to True to write the built dict to file. Default is None
    output: str
        output object
    Returns
    -------
    dict:
        return a dict of built Instrument classes

    Examples
    --------
    >>> buildChannels = BuildChannels()
    >>> channels = buildChannels(payload=payload, write=False, output=None)
    """

    def __init__(self):
        self.addTaskParam('payload', 'main payload')
        self.addTaskParam('write', 'write processed instrument to output file')
        self.addTaskParam('output', 'output object')

    def execute(self):
        self.info('building channel')
        channels = {}
        self.debug('detectors found : {}'.format(
            self.get_task_param('payload')['channel'].keys()))

        ch = None
        if self.get_task_param('write'):
            inst = self.get_task_param('output').create_group('payload')
            inst.store_dictionary(self.get_task_param('payload'),
                                  group_name='payload description')
            ch = inst.create_group('channels')

        buildInstrument = BuildInstrument()
        for det in self.get_task_param('payload')['channel'].keys():
            channel_type = \
            self.get_task_param('payload')['channel'][det]['channelClass'][
                'value'].lower()
            channels[det] = buildInstrument(type=channel_type, name=det,
                                            description=
                                            self.get_task_param('payload')[
                                                'channel'][det],
                                            payload=self.get_task_param(
                                                'payload'),
                                            write=False,
                                            output=None)
            if self.get_task_param('write'):
                channels[det].write(ch)
        self.debug('channels : {}'.format(channels))
        self.set_output(channels)


class LoadPayload(Task):
    """
    Loads payload and channels from dict

    Parameters
    ----------
    input: dict
        main dictionary


    Returns
    -------
    dict:
        payload dict
    dict:
        return a dict of built Instrument classes

    Examples
    --------
    >>> loadPayload = LoadPayload()
    >>> channels = loadPayload(input=input)
    """

    def __init__(self):
        self.addTaskParam('input', 'input data')

    def execute(self):
        payload_dir = self.get_task_param('input')['payload']
        payload = load(payload_dir['payload description'])
        channels_dir = payload_dir['channels']
        channels = {}
        for ch in channels_dir.keys():
            ch_dir = channels_dir[ch]
            description = load(ch_dir['description'])
            instrument = instruments[
                description['channelClass']['value'].lower().decode("utf-8")]
            channels[ch] = instrument(name=ch,
                                      description=description,
                                      payload=payload,
                                      )
            table = read_table_hdf5(ch_dir, path=ch)
            built_instr = load(ch_dir['built_instr'])
            channels[ch].load(table, built_instr)
        self.debug('channels loaded: {}'.format(channels))
        self.set_output([payload, channels])


class PreparePayload(Task):
    """
    It loads the payload information and returns a dictionary containing the payload,
    a dictionary with the channels already built,
    and a couple with minimum and maximum wavelength investigated by the payload

    Parameters
    ----------
    payload_file: str
        xml file with payload description
    output: str
        h5 output file

    Returns
    -------

    dict:
        payload dictionary
    dict:
        channels dictionary
    couple:
        minimum and maximum investigated wavelength

    Examples
    --------
    >>> preparePayload = PreparePayload()
    >>> payload, channels, (wl_min, wl_max) = PreparePayload(payload_file=payload, output=output)

    """

    def __init__(self):
        self.addTaskParam('payload_file', 'payload xml file')
        self.addTaskParam('output', 'output file')

    def execute(self):
        import os
        from exorad.tasks import LoadOptions
        from exorad.output.hdf5 import HDF5Output

        loadOptions = LoadOptions()
        buildChannels = BuildChannels()
        loadPayload = LoadPayload()

        payload_file = self.get_task_param('payload_file')
        output = self.get_task_param('output')

        if isinstance(payload_file, str):
            ext = os.path.splitext(payload_file)[1]
        else:
            ext = None

        if (ext == '.xml') or isinstance(payload_file, dict):
            if (ext == '.xml'):
                payload = loadOptions(filename=payload_file)
            if isinstance(payload_file, dict):
                payload = payload_file
            if output is not None:
                with HDF5Output(output) as out:
                    channels = buildChannels(payload=payload, write=True,
                                             output=out)
            else:
                channels = buildChannels(payload=payload, write=False,
                                         output=None)

        elif ext == 'h5':
            payload, channels = loadPayload(input=payload_file)
        else:
            self.error('Unsupported payload format')
            raise IOError('Unsupported payload format')
        wl_min, wl_max = payload['common']['wl_min']['value'], \
                         payload['common']['wl_max']['value']
        self.set_output([payload, channels, (wl_min, wl_max)])


class MergeChannelsOutput(Task):
    """
    Merges the channels output tables

    Parameters
    ----------
    channels: dict

    Returns
    ---------
    Table
    """

    def __init__(self):
        self.addTaskParam('channels', 'dict of channels')

    def execute(self):
        from exorad.utils.util import vstack_tables
        channels = self.get_task_param('channels')
        table_list = []
        for channel in channels.keys():
            table_list.append(channels[channel].table)
        table = vstack_tables(table_list)
        self.set_output(table)


class GetChannelList(Task):
    """
    Returns a list of the channels of the specific type that are contained in the instrument options file

    Parameters
    ----------
    options: object
        payload description
    channel_type: string
        type of channel to find

    Returns
    -------
    list:
        list of strings with channel names

    Examples
    --------
    >>> from exorad.tasks.loadOptions import LoadOptions
    >>> loadOptions = LoadOptions()
    >>> options = loadOptions(filename = 'path/to/file.xml')
    >>> getDetectorList = GetChannelList()
    >>> photometers_list = getDetectorList(options = options, channel_type = 'Photometer')
    >>> spectrometers_list = getDetectorList(options = options, channel_type = 'Spectrometer')
    """

    def __init__(self):
        self.addTaskParam('options', 'instrument description class')
        self.addTaskParam('channel_type', 'desired channel type')

    def execute(self):
        self._opt = self.get_task_param('options')
        self._channel_type = self.get_task_param('channel_type')
        self.channel_lst = self.__get_channel_list__()
        self.info('{} type channel found: {}'.format(self._channel_type,
                                                     self.channel_lst))
        self.set_output(self.channel_lst)

    def __get_channel_list__(self):
        channel_lst = []
        for channel in self._opt['channel'].keys():
            if self._opt['channel'][channel]['channelClass'][
                'value'].lower() == self._channel_type.lower():
                channel_lst.append(channel)
                self.debug(
                    '{} added as {}'.format(channel, self._channel_type))
        return channel_lst



import numpy as np
from astropy import units as u

from exorad.models.source import Star, CustomSed
from exorad.tasks.task import Task
import os

class LoadSource(Task):
    """
    Updates target with its source and return the source Sed

    Parameters
    ----------
    target: Target Class
        target class
    source: dict
        source spectrum description
    wl_range: couple
        wavelength range to investigate: (wl_min, wl_max)


    Returns
    -------
        Target:
            Target class with star.sed updated
        Sed:
            source Sed

    Examples
    --------
    >>> loadSource = LoadSource()
    >>> target, source = loadSource(target= target, source={'sourceSpectrum': {'value':'Planck'}})

    Raises
    ------
    AttributeError:
        if some target information are missing
    """

    def __init__(self):
        self.addTaskParam('target', 'target class object')
        self.addTaskParam('source', 'source spectrum description')
        self.addTaskParam('wl_range', 'wavelength range to investigate')

    def execute(self):
        target = self.get_task_param('target')
        source = self.get_task_param('source')

        if isinstance(source, str):
            source = {'value': source}
            self.warning('source should be dict, not string.')

        if source['value'].lower() == 'custom':
            # if custom source, only R and D are needed for the solid angle
            for attr in ['D', 'R']:
                if not hasattr(target.star.__getattribute__(attr), 'value'):
                    self.error('target information incomplete')
                    raise AttributeError('target information incomplete')
            star = CustomSed(source['CustomSed']['value'],
                             target.star.R,
                             target.star.D)
            self.debug('custom sed used {}'.format(source['CustomSed']['value']))

        else:
            # check if star information are complete
            for attr in ['D', 'Teff', 'M', 'R']:
                if not hasattr(target.star.__getattribute__(attr), 'value'):
                    self.error('target information incomplete')
                    raise AttributeError('target information incomplete')

            self.debug('source spectrum : {}'.format(source['value'].lower()))
            if source['value'].lower() == 'planck':
                self.debug('Plack sed selected')
                star = Star('.',
                            target.star.D,
                            target.star.Teff,
                            target.star.calc_logg(target.star.M, target.star.R),
                            0.0,
                            target.star.R,
                            use_planck_spectrum=True)

            elif source['value'].lower() == 'phoenix':
                try:
                    star_sed_path = source['StellarModels']['value']
                except KeyError:
                    if os.environ.get('PHOENIX_PATH', None) is not None:
                        star_sed_path = os.environ.get('PHOENIX_PATH', None)
                    else:
                        raise IOError('No phoenix path specificed')

                if not os.path.exists(star_sed_path):
                    raise IOError('Phoenix path does not exist: {}'.format(star_sed_path))
                    
                star = Star(star_sed_path,
                            target.star.D,
                            target.star.Teff,
                            target.star.calc_logg(target.star.M, target.star.R),
                            0.0,
                            target.star.R,
                            use_planck_spectrum=False)
                self.debug('stellar sed used {}'.format(star.filename))
            else:
                star = Star('.',
                            target.star.D,
                            target.star.Teff,
                            target.star.calc_logg(target.star.M, target.star.R),
                            0.0,
                            target.star.R,
                            use_planck_spectrum=True)
                self.info('invalid source spectrum description. Planck spectrum is used')

        wl_min, wl_max = self.get_task_param('wl_range')
        if wl_min > wl_max: wl_min, wl_max = wl_max, wl_min
        if not hasattr(wl_min, 'unit'):
            self.debug('wavelength unit not found: micron assumed')
            wl_min *= u.um
            wl_max *= u.um

        wl_grid = np.logspace(np.log10(wl_min.value),
                              np.log10(wl_max.value), 6000) * \
                  wl_max.unit
        star.sed.spectral_rebin(wl_grid)

        target.update_target(star)
        if hasattr(target, 'table'):
            target.table = self._add_star_metadata(target.star, target.table)
        self.set_output([target, star.sed])

    def _add_star_metadata(self, star, table):
        metadata = {}

        metadata['starName'] = star.name
        metadata['starM'] = star.M
        metadata['starTeff'] = star.Teff
        metadata['starR'] = star.R
        metadata['starDistance'] = star.D
        metadata['starL'] = star.luminosity
        metadata['starModel'] = star.model
        if hasattr(star, 'magk'):
            metadata['starMagK'] = star.magk

        table.meta.update(metadata)
        return table



from .singleton import Singleton
from exorad.log import Logger


class GlobalCache(Singleton, Logger):
    """

    Allows for the storage of global variables

    """

    def init(self):
        super().__init__()
        self.variable_dict = {}

    def __getitem__(self, key):
        try:
            return self.variable_dict[key]
        except KeyError:
            return None

    def __setitem__(self, key, value):
        self.variable_dict[key] = value


import logging
from collections import OrderedDict

import numpy as np
from astropy import units as u

from exorad.models.signal import Signal, CustomSignal

logger = logging.getLogger('exorad.noise')


def frame_time(target, channel, out):
    """
    Given the channel and channel descriptions, populates the output table with saturation and frame times

    Parameters
    -----------
    channel: dict
        channel description
    target: Target
        Target to investigate
    out: QTable
        output table


    Returns
    --------
    QTable
        output table populated
    """
    name = channel['value']
    max_signal_in_pix = target.table['MaxSignal_inPixel'][target.table['chName'] == name]
    out['saturation_time'] = channel['detector']['well_depth']['value'] / max_signal_in_pix
    logger.debug('saturation time : {}'.format(out['saturation_time']))
    if 'frame_time' in channel['detector'].keys():
        out['frameTime'] = channel['detector']['frame_time']['value'] * np.ones(out['saturation_time'].size)
    else:
        out['frameTime'] = channel['detector']['f_well_depth']['value'] * np.min(out['saturation_time']) \
                           * np.ones(out['saturation_time'].size)
    logger.debug('frame time : {}'.format(out['frameTime']))
    return out


def multiaccum(channel, t_frame):
    """
    Given the channel and time frame, returns the multiaccum estimation for read and shot gain

    Parameters
    -----------
    channel: dict
        channel description
    t_frame: float
        frame time

    Returns
    --------
    float
        read noise gain
    float
        shot noise gain
    """
    nRead = np.floor(t_frame * channel['detector']['freqNDR']['value'])

    if 'multiaccumM' in channel['detector']:
        m = channel['detector']['multiaccumM']['value']
        logger.debug('multiaccum activated: m = {}'.format(m))
        tf = 0. * u.s
    else:
        m = 1
        tf = 0. * u.s

    if nRead < 2:
        nRead = 2.0  # Force to CDS in nRead < 2

    read_gain = 12.0 * (nRead - 1.0) / (nRead ** 2 + nRead) / m
    shot_gain = 6.0 * (nRead ** 2 + 1.0) / (nRead ** 2 + nRead) / 5.0 * \
                (1 - 5. / 3. * (m ** 2 - 1) / m / (nRead ** 2 + 1) * tf / (nRead - 1) / t_frame)
    logger.debug('read noise gain: {}'.format(read_gain))
    logger.debug('shot noise gain: {}'.format(shot_gain))
    return read_gain, shot_gain


def photon_noise(table, channel, shot_gain, out):
    """
    Given the channel and channel descriptions, populates the output table with photon noises

    Parameters
    -----------
    channel: dict
        channel description
    target: Table
        Target table to investigate
    shot_gain: float
        multiaccum factor for photon noise
    out: QTable
        output table

    Returns
    --------
    QTable
        output table populated
    """
    name = channel['value']

    signals = [key for key in table.keys() if 'signal' in key]
    for key in signals:
        noise_key = '{}_noise'.format(key)
        out[noise_key] = np.sqrt(shot_gain * table[key][table['chName'] == name] * u.count / u.hr).to(
            u.count / u.s) * u.hr ** 0.5
        logger.debug('{} : {}'.format(noise_key, out[noise_key]))
    return out

def photon_noise_variance(table, out):
    '''
    """
    Given the target table return the sum of photon variances

    Parameters
    -----------
    target: Table
        Target table to investigate
    out: QTable
        output table

    Returns
    --------
    Array
        sum of photon variances
    """
    '''
    signals = [key for key in table.keys() if 'signal' in key]
    _photon_noise_variance = np.zeros(out['frameTime'].size) * (u.count / u.s) ** 2 * u.hr
    for key in signals:
        noise_key = '{}_noise'.format(key)
        _photon_noise_variance += out[noise_key] * out[noise_key]
    return _photon_noise_variance


def add_custom_noise(custom, wl, out):
    if 'data' in custom:
        col_name = [col for col in custom['data'].keys() if 'Wavelength' not in col]
        custom_noise = Signal(custom['data']['Wavelength'],
                              custom['data'][col_name])
        custom_noise.spectral_rebin(wl)
        out['{}_noise'.format(col_name)] = custom_noise.data
        logger.debug('{} added as custom noise :{}'.format(col_name, custom_noise.data))
        out['total_noise'] = np.sqrt(out['total_noise'] * out['total_noise'] +
                                     custom_noise.data * custom_noise.data)

    if isinstance(custom, OrderedDict):
        for contrib in custom:
            custom_noise = custom[contrib]['value'] * 1e-6 * np.ones(wl.size) * u.hr ** 0.5
            out['{}_noise'.format(custom[contrib]['name']['value'])] = custom_noise
            logger.debug('{} added as custom noise :{}'.format(custom[contrib]['name']['value'],
                                                               custom_noise))
            out['total_noise'] = np.sqrt(out['total_noise'] * out['total_noise'] +
                                         custom_noise * custom_noise)
    else:
        custom_noise = custom['value'] * 1e-6 * np.ones(wl.size) * u.hr ** 0.5
        out['{}_noise'.format(custom['name']['value'])] = custom_noise
        logger.debug('{} added as custom noise :{}'.format(custom['name']['value'],
                                                           custom_noise))
        out['total_noise'] = np.sqrt(out['total_noise'] * out['total_noise'] +
                                     custom_noise * custom_noise)
    return out

class Noise(CustomSignal):
    """
    It's a Signal class with data having units of [hr^1/2]
    """

    def __init__(self, wl_grid, data, time_grid=[0] * u.hr):
        super().__init__(wl_grid, data, u.hr ** 0.5, time_grid)



import logging

from .logger import Logger

last_log = logging.INFO


def setLogLevel(level):
    global last_log
    from .logger import root_logger
    root_logger.handlers[0].setLevel(level)
    #    root_logger.setLevel(level)
    last_log = level


def disableLogging():
    # import logging
    # from .logger import root_logger
    # root_logger.setLevel(logging.ERROR)
    setLogLevel(logging.ERROR)

def enableLogging():
    # global last_log
    # import logging
    # from .logger import root_logger
    # root_logger.setLevel(logging.INFO)
    #
    # # if last_log is None:
    # #     last_log = logging.INFO
    setLogLevel(logging.INFO)


def addLogFile(fname='exorad.log'):
    from .logger import root_logger
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    file_handler = logging.FileHandler(fname)
    file_handler.setFormatter(formatter)
    file_handler.setLevel(logging.DEBUG)
    root_logger.addHandler(file_handler)



import os
import xml.etree.ElementTree as ET
from collections import OrderedDict

from astropy import units as u
from astropy.io.ascii.core import InconsistentTableError
from astropy.io import ascii

from .task import Task

compactString = lambda string: string.replace('\n', '').strip()


class LoadOptions(Task):
    """
    Reads the xml file with payload description and return an object with attributes related to the input data

    Parameters
    ----------
    filename: string
        input xml file location
    config_path: string (optional)
        on-run setting for ConfigPat. Default is None.

    Returns
    -------
    dict:
        parsed xml input file

    Raises
    ------
        IOError
            if the indicated file is not found or the format is not supported

    Examples
    --------
    >>> loadOptions = LoadOptions()
    >>> options = loadOptions(filename = 'path/to/file.xml')
    """

    def __init__(self):
        self.addTaskParam('filename', 'input option file name')
        self.addTaskParam('config_path', 'on-run setting for ConfigPath', None)

    def execute(self):
        self._filename = self.get_task_param('filename')
        self._config_path = self.get_task_param('config_path')
        self.__check_format__()
        root = self.__get_root__()
        self._dict = self.__parser__(root)
        # self.debug('loaded options: {}'.format(self._dict))
        self.set_output(self._dict)

    configPath = None

    def __check_format__(self):
        if not self._filename.endswith('.xml'):
            self.error('wrong input format')
            raise IOError

    def __get_root__(self):
        try:
            self.debug('input option file found %s' % self._filename)
            return ET.parse(self._filename).getroot()
        except IOError:
            self.error('No input option file found')
            raise IOError

    def __parser__(self, root):
        root_dict = {}

        for ch in root:

            retval = self.__parser__(ch)
            # parse all attributes
            for key in list(ch.attrib.keys()):
                retval[key] = ch.attrib[key]

            value = compactString(ch.text)
            if (value is not None) and (value != ''):
                try:
                    value = int(value)
                except ValueError:
                    if value.replace('.', ''):
                        try:
                            value = float(value)
                        except ValueError:
                            pass
                if 'unit' in retval:
                    unitName = retval['unit']
                    if unitName == 'dimensionless': unitName = ''
                    value = value * u.Unit(unitName)
                if value == 'True':
                    value = bool(True)
                if value == 'False':
                    value = bool(False)
                if isinstance(value, str):
                    if '__ConfigPath__' in value:
                        value = value.replace('__ConfigPath__', self.configPath)
                if ch.tag == "ConfigPath":
                    self.configPath = value

                if self._config_path:
                    self.configPath = self._config_path

                # if isinstance(value, str):
                #     retval = value
                # else:
                retval['value'] = value

            if ch.tag in root_dict:
                # if an other instance of same tag exists, transform into a dict
                attr = root_dict[ch.tag]
                if isinstance(attr, OrderedDict):
                    attr[value] = retval
                else:
                    if not isinstance(attr, str):
                        dtmp = OrderedDict([(attr['value'], attr), (value, retval)])
                        root_dict[ch.tag] = dtmp
            else:
                # othewise, set new attr
                root_dict[ch.tag] = retval

        if 'datafile' in root_dict:
            try:
                datatype_attr = root_dict['datatype']
                datatype = datatype_attr['value']
            except:
                datatype = 'ecsv'

            attrValue = root_dict['datafile']
            datafile = attrValue
            datafile = datafile['value'].replace('__ConfigPath__', self.configPath)
            try:
                data = self.__read_datatable__(datafile, datatype)
                root_dict['data'] = data
            except IOError:
                self.error("Cannot read input file")
                raise IOError

        if 'datadict' in root_dict:
            attrValue = root_dict['datadict']
            datafile = attrValue
            datafile = datafile['value'].replace('__ConfigPath__', self.configPath)
            try:
                data = self.__read_datadict__(datafile)
                root_dict['data'] = data
            except IOError:
                self.error("Cannot read input file")
                raise IOError
        if 'config_dict' in root_dict:
            attrValue = root_dict['config_dict']
            datafile = attrValue
            datafile = datafile['value'].replace('__ConfigPath__', self.configPath)
            try:
                data = self.__read_datadict__(datafile)
                root_dict = data
                root_dict['config_dict'] = datafile
                self.debug('Configuration dictionary found in {}'.format(datafile))
            except IOError:
                self.error("Cannot read input file")
                raise IOError

        return root_dict

    def getopt(self):
        return self.__obj__

    def __read_datatable__(self, datafile, datatype):
        if datatype == 'ecsv':
            try:
                data = ascii.read(os.path.expanduser(datafile),
                                  format='ecsv')
            except InconsistentTableError:
                data = ascii.read(os.path.expanduser(datafile),
                                  format='csv')
            return data

        else:
            self.error('inconsistent table')
            raise IOError

    def __read_datadict__(self, datadict):
        ext = os.path.splitext(datadict)[1]
        if ext in ['.h5', '.hdf5']:
            from exorad.output.hdf5 import load
            data = load(datadict)
            return data
        elif ext == '.pickle':
            import pickle
            f = open(datadict, 'rb')
            data = pickle.load(f)
            f.close()
            return data
        else:
            self.error('inconsistent table')
            raise IOError

